@@ -19,6 +19,7 @@ import os
 
 from oslo_concurrency import processutils as putils
 from oslo_log import log as logging
+import six
 
 from os_brick.initiator import linuxscsi
 
@@ -34,7 +35,7 @@ class LinuxFibreChannel(linuxscsi.LinuxSCSI):
         else:
             return False
 
-    def _get_hba_channel_scsi_target(self, hba):
+    def _get_hba_channel_scsi_target(self, hba, conn_props):
         """Try to get the HBA channel and SCSI target for an HBA.
 
         This method only works for Fibre Channel targets that implement a
@@ -43,31 +44,71 @@ class LinuxFibreChannel(linuxscsi.LinuxSCSI):
 
         :returns: List or None
         """
+        # We want the target's WWPNs, so we use the initiator_target_map if
+        # present for this hba or default to target_wwns if not present.
+        wwpns = conn_props['target_wwn']
+        if 'initiator_target_map' in conn_props:
+            wwpns = conn_props['initiator_target_map'].get(hba['port_name'],
+                                                           wwpns)
+
+        # If it's not a string then it's an iterable (most likely a list),
+        # so we need to create a BRE for the grep query.
+        if not isinstance(wwpns, six.string_types):
+            wwpns = '\|'.join(wwpns)
+
         # Leave only the number from the host_device field (ie: host6)
         host_device = hba['host_device']
         if host_device and len(host_device) > 4:
             host_device = host_device[4:]
 
         path = '/sys/class/fc_transport/target%s:' % host_device
-        cmd = 'grep %(wwnn)s %(path)s*/node_name' % {'wwnn': hba['node_name'],
-                                                     'path': path}
+        # Since we'll run the command in a shell ensure BRE are being used
+        cmd = 'grep -Gil "%(wwpns)s" %(path)s*/port_name' % {'wwpns': wwpns,
+                                                             'path': path}
         try:
-            out, _err = self._execute(cmd)
+            # We need to run command in shell to expand the * glob
+            out, _err = self._execute(cmd, shell=True)
             return [line.split('/')[4].split(':')[1:]
                     for line in out.split('\n') if line.startswith(path)]
         except Exception as exc:
             LOG.debug('Could not get HBA channel and SCSI target ID, path: '
-                      '%(path)s, reason: %(reason)s', {'path': path,
-                                                       'reason': exc})
+                      '%(path)s*, reason: %(reason)s', {'path': path,
+                                                        'reason': exc})
             return None
 
-    def rescan_hosts(self, hbas, target_lun):
-        for hba in hbas:
-            # Try to get HBA channel and SCSI target to use as filters
-            cts = self._get_hba_channel_scsi_target(hba)
-            # If we couldn't get the channel and target use wildcards
-            if not cts:
-                cts = [('-', '-')]
+    def rescan_hosts(self, hbas, connection_properties):
+        LOG.debug('Rescaning HBAs %(hbas)s with connection properties '
+                  '%(conn_props)s', {'hbas': hbas,
+                                     'conn_props': connection_properties})
+        target_lun = connection_properties['target_lun']
+        get_cts = self._get_hba_channel_scsi_target
+
+        # Use initiator_target_map provided by backend as HBA exclussion map
+        ports = connection_properties.get('initiator_target_map')
+        if ports:
+            hbas = [hba for hba in hbas if hba['port_name'] in ports]
+            LOG.debug('Using initiator target map to exclude HBAs')
+            process = [(hba, get_cts(hba, connection_properties))
+                       for hba in hbas]
+
+        # With no target map we'll check if target implements single WWNN for
+        # all ports, if it does we only use HBAs connected (info was found),
+        # otherwise we are forced to blindly scan all HBAs.
+        else:
+            with_info = []
+            no_info = []
+
+            for hba in hbas:
+                cts = get_cts(hba, connection_properties)
+                target_list = with_info if cts else no_info
+                cts = cts or [('-', '-')]
+                target_list.append((hba, cts))
+
+            process = with_info or no_info
+            msg = "implements" if with_info else "doesn't implement"
+            LOG.debug('FC target %s single WWNN for all ports.', msg)
+
+        for hba, cts in process:
             for hba_channel, target_id in cts:
                 LOG.debug('Scanning host %(host)s (wwnn: %(wwnn)s, c: '
                           '%(channel)s, t: %(target)s, l: %(lun)s)',
@@ -264,54 +305,3 @@ class LinuxFibreChannelS390X(LinuxFibreChannel):
             LOG.warning("unit_remove call for s390 failed exit %(code)s, "
                         "stderr %(stderr)s",
                         {'code': exc.exit_code, 'stderr': exc.stderr})
-
-
-class LinuxFibreChannelPPC64(LinuxFibreChannel):
-
-    def _get_hba_channel_scsi_target(self, hba, wwpn):
-        """Try to get the HBA channel and SCSI target for an HBA.
-
-        This method works for Fibre Channel targets iterating over all the
-        target wwpn port and finding the c, t, l. so caller should expect us to
-        return either None or an empty list.
-        """
-        # Leave only the number from the host_device field (ie: host6)
-        host_device = hba['host_device']
-        if host_device and len(host_device) > 4:
-            host_device = host_device[4:]
-        path = '/sys/class/fc_transport/target%s:' % host_device
-        cmd = 'grep -il %(wwpn)s %(path)s*/port_name' % {'wwpn': wwpn,
-                                                         'path': path}
-        try:
-            out, _err = self._execute(cmd, shell=True)
-            return [line.split('/')[4].split(':')[1:]
-                    for line in out.split('\n') if line.startswith(path)]
-        except Exception as exc:
-            LOG.error("Could not get HBA channel and SCSI target ID, "
-                      "reason: %s", exc)
-            return None
-
-    def rescan_hosts(self, hbas, target_lun):
-        for hba in hbas:
-            # Try to get HBA channel and SCSI target to use as filters
-            # Ignore HBA which does not have target wwn
-            if 'target_wwn' not in hba.keys():
-                continue
-            for wwpn in hba['target_wwn']:
-                cts = self._get_hba_channel_scsi_target(hba, wwpn)
-                # If we couldn't get the channel and target use wildcards
-                if not cts:
-                    cts = [('-', '-')]
-                for hba_channel, target_id in cts:
-                    LOG.debug('Scanning host %(host)s (wwpn: %(wwpn)s, c: '
-                              '%(channel)s, t: %(target)s, l: %(lun)s)',
-                              {'host': hba['host_device'],
-                               'wwpn': hba['target_wwn'],
-                               'channel': hba_channel,
-                               'target': target_id,
-                               'lun': target_lun})
-                    self.echo_scsi_command(
-                        "/sys/class/scsi_host/%s/scan" % hba['host_device'],
-                        "%(c)s %(t)s %(l)s" % {'c': hba_channel,
-                                               't': target_id,
-                                               'l': target_lun})
